"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const core_container_1 = require("@mlcc/core-container");
const core_interfaces_1 = require("@mlcc/core-interfaces");
const core_utils_1 = require("@mlcc/core-utils");
const fs_1 = __importDefault(require("fs"));
const chunk_1 = __importDefault(require("lodash/chunk"));
const path_1 = __importDefault(require("path"));
const pg_promise_1 = __importDefault(require("pg-promise"));
const migrations_1 = require("./migrations");
const repositories_1 = require("./repositories");
const spv_1 = require("./spv");
const query_executor_1 = require("./sql/query-executor");
const utils_1 = require("./utils");
class PostgresConnection {
    constructor(options, walletManager) {
        this.options = options;
        this.walletManager = walletManager;
        this.logger = core_container_1.app.resolvePlugin("logger");
        this.models = {};
        this.emitter = core_container_1.app.resolvePlugin("event-emitter");
    }
    async buildWallets(height) {
        const spvPath = `${process.env.CORE_PATH_CACHE}/spv.json`;
        if (fs_1.default.existsSync(spvPath)) {
            fs_1.default.removeSync(spvPath);
            this.logger.info("MLC Core ended unexpectedly - resuming from where we left off :runner:");
            return true;
        }
        try {
            const spv = new spv_1.SPV(this.query, this.walletManager);
            return await spv.build(height);
        }
        catch (error) {
            this.logger.error(error.stack);
        }
        return false;
    }
    async commitQueuedQueries() {
        if (!this.queuedQueries || this.queuedQueries.length === 0) {
            return;
        }
        this.logger.debug("Committing database transactions.");
        try {
            await this.db.tx(t => t.batch(this.queuedQueries));
        }
        catch (error) {
            this.logger.error(error);
            throw error;
        }
        finally {
            this.queuedQueries = null;
        }
    }
    async connect() {
        this.emitter.emit(core_interfaces_1.Database.DatabaseEvents.PRE_CONNECT);
        const initialization = {
            receive(data, result, e) {
                utils_1.camelizeColumns(pgp, data);
            },
            extend(object) {
                for (const repository of Object.keys(repositories_1.repositories)) {
                    object[repository] = new repositories_1.repositories[repository](object, pgp);
                }
            },
        };
        const pgp = pg_promise_1.default(Object.assign({}, this.options.initialization, initialization));
        this.pgp = pgp;
        this.db = this.pgp(this.options.connection);
    }
    async deleteBlock(block) {
        try {
            const queries = [this.transactionsRepository.deleteByBlockId(block.data.id), this.blocksRepository.delete(block.data.id)];
            await this.db.tx(t => t.batch(queries));
        }
        catch (error) {
            this.logger.error(error.stack);
            throw error;
        }
    }
    async disconnect() {
        this.logger.debug("Disconnecting from database");
        this.emitter.emit(core_interfaces_1.Database.DatabaseEvents.PRE_DISCONNECT);
        try {
            await this.commitQueuedQueries();
            this.cache.clear();
        }
        catch (error) {
            this.logger.warn("Issue in commiting blocks, database might be corrupted");
            this.logger.warn(error.message);
        }
        await this.pgp.end();
        this.emitter.emit(core_interfaces_1.Database.DatabaseEvents.POST_DISCONNECT);
        this.logger.debug("Disconnected from database");
    }
    enqueueDeleteBlock(block) {
        const queries = [this.transactionsRepository.deleteByBlockId(block.data.id), this.blocksRepository.delete(block.data.id)];
        this.enqueueQueries(queries);
    }
    enqueueDeleteRound(height) {
        const { round, nextRound, maxDelegates } = core_utils_1.roundCalculator.calculateRound(height);
        if (nextRound === round + 1 && height >= maxDelegates) {
            this.enqueueQueries([this.roundsRepository.delete(nextRound)]);
        }
    }
    enqueueSaveBlock(block) {
        const queries = [this.blocksRepository.insert(block.data)];
        if (block.transactions.length > 0) {
            queries.push(this.transactionsRepository.insert(block.transactions));
        }
        this.enqueueQueries(queries);
    }
    async make() {
        if (this.db) {
            throw new Error("Database connection already initialised");
        }
        this.logger.debug("Connecting to database");
        this.queuedQueries = null;
        this.cache = new Map();
        try {
            await this.connect();
            this.exposeRepositories();
            await this.registerQueryExecutor();
            await this.runMigrations();
            await this.registerModels();
            this.logger.debug("Connected to database.");
            this.emitter.emit(core_interfaces_1.Database.DatabaseEvents.POST_CONNECT);
            return this;
        }
        catch (error) {
            core_container_1.app.forceExit("Unable to connect to the database!", error);
        }
        return null;
    }
    async saveBlock(block) {
        try {
            const queries = [this.blocksRepository.insert(block.data)];
            if (block.transactions.length > 0) {
                queries.push(this.transactionsRepository.insert(block.transactions));
            }
            await this.db.tx(t => t.batch(queries));
        }
        catch (err) {
            this.logger.error(err.message);
        }
    }
    async saveWallets(wallets, force) {
        if (force) {
            // all wallets to be updated, performance is better without upsert
            await this.walletsRepository.truncate();
            try {
                const chunks = chunk_1.default(wallets, 5000).map(c => this.walletsRepository.insert(c)); // this 5000 figure should be configurable...
                await this.db.tx(t => t.batch(chunks));
            }
            catch (error) {
                this.logger.error(error.stack);
            }
        }
        else {
            // NOTE: The list of delegates is calculated in-memory against the WalletManager,
            // so it is safe to perform the costly UPSERT non-blocking during round change only:
            // 'await saveWallets(false)' -> 'saveWallets(false)'
            try {
                const queries = wallets.map(wallet => this.walletsRepository.updateOrCreate(wallet));
                await this.db.tx(t => t.batch(queries));
            }
            catch (error) {
                this.logger.error(error.stack);
            }
        }
    }
    /**
     * Run all migrations.
     * @return {void}
     */
    async runMigrations() {
        for (const migration of migrations_1.migrations) {
            const { name } = path_1.default.parse(migration.file);
            if (name === "20180304100000-create-migrations-table") {
                await this.query.none(migration);
            }
            else {
                const row = await this.migrationsRepository.findByName(name);
                if (row === null) {
                    this.logger.debug(`Migrating ${name}`);
                    await this.query.none(migration);
                    await this.migrationsRepository.insert({ name });
                }
            }
        }
    }
    /**
     * Register all models.
     * @return {void}
     */
    async registerModels() {
        for (const [key, Value] of Object.entries(require("./models"))) {
            this.models[key.toLowerCase()] = new Value(this.pgp);
        }
    }
    /**
     * Register the query builder.
     * @return {void}
     */
    registerQueryExecutor() {
        this.query = new query_executor_1.QueryExecutor(this);
    }
    enqueueQueries(queries) {
        if (!this.queuedQueries) {
            this.queuedQueries = [];
        }
        this.queuedQueries.push(...queries);
    }
    exposeRepositories() {
        this.blocksRepository = this.db.blocks;
        this.transactionsRepository = this.db.transactions;
        this.roundsRepository = this.db.rounds;
        this.walletsRepository = this.db.wallets;
        this.migrationsRepository = this.db.migrations;
    }
}
exports.PostgresConnection = PostgresConnection;
//# sourceMappingURL=postgres-connection.js.map